#include <stdio.h>
#include <stdlib.h>

#define READ_ERROR       0
#define WRITE_ERROR      1
#define MEMORY           2
#define WRONG_FORMAT     3
#define DEGREE_OVERFLOW  4

#define LEFT 0
#define RIGHT 1

#define SPACEDIM 3

/* Algorithme en 2 temps */
/* 1) pour chaque sommet, on construit une liste des artes adjacentes */
/*    a ce sommet. Chaque arte contient les indices de ces 2 sommets, */
/*    le premier tant toujours plus petit que le second, ainsi que les */
/*    indices des faces  gauche de l'arte et  droite de l'arte */
/* */
/* 2) on tourne autour de chaque sommet, en connectant les artes adjacentes */
/*    dans le bon sens */

typedef struct s_Vertex{
           double V[SPACEDIM];
           double N[SPACEDIM];
           int    F;
                     } Vertex;

typedef struct s_Face{
           int    nVert;
           int    *V;
           int    *F;
           double N[3];
                     } Face;

typedef struct s_Edge{
           int    V[2];
           int    F[2];
           struct s_Edge *Next;
                     } Edge;

void PrintError( int ErrorCode, char *origin)
  {
  char message[256];

  switch(ErrorCode)
    {
    case READ_ERROR:
            sprintf( message, "file %s can not be read\n", origin);
            break;
    case WRITE_ERROR:
            sprintf( message, "file %s can not be writen\n", origin);
            break;
    case MEMORY:
            sprintf( message, "not enough memory to allocate %s\n", origin);
            break;
    case WRONG_FORMAT:
            sprintf( message, "wrong format for %s\n", origin);
            break;
    }

  fprintf( stderr, "%s", message);
  exit(-1);
  }

void ReadOff( FILE *fin, int *pn_vertex, int *pn_face, Vertex **plist_v, Face **plist_f)
  {
  char format[100];
  Vertex *list_v;
  Face *list_f;
  int n_vertex, n_face, dummy, fi, ei;

  fscanf( fin, "%s\n", format);

  if (strcmp( format, "OFF") != 0) {PrintError( WRONG_FORMAT, "input file: must be in OFF format");}

  fscanf( fin, "%d %d %d\n", pn_vertex, pn_face, &dummy);

  n_vertex = *pn_vertex;
  n_face = *pn_face;

  *plist_v = (Vertex*) malloc(sizeof(Vertex) * n_vertex);
  list_v = *plist_v;
  if (list_v==NULL) {PrintError( MEMORY, "vertices");}

  *plist_f = (Face*) malloc(sizeof(Face) * n_face);
  list_f = *plist_f;
  if (list_f==NULL) {PrintError( MEMORY, "faces");}

  for (fi=0; fi<n_vertex; fi++)
    {
    fscanf( fin, "%lf %lf %lf",
                  &list_v[fi].V[0], &list_v[fi].V[1], &list_v[fi].V[2]);
    }

  for (fi=0; fi<n_face; fi++) {

    fscanf( fin, "%d", &list_f[fi].nVert);
    fprintf( stderr, "%d ", list_f[fi].nVert);

    list_f[fi].V = (int*) malloc(sizeof(int) * list_f[fi].nVert);
    list_f[fi].F = (int*) malloc(sizeof(int) * list_f[fi].nVert);

    for (ei=0; ei<list_f[fi].nVert; ei++)
      {
      fscanf( fin, "%d", &list_f[fi].V[ei]);
      fprintf( stderr, "%03d ", list_f[fi].V[ei]);
      }
    fprintf( stderr, "\n");

  }

  fprintf( stderr, "Read OFF file with %d vertices and %d faces\n", n_vertex, n_face);
  }

void WriteGFile(FILE *fout, int n_vertex, int n_face, Vertex *list_v, Face *list_f)
  {
  int i, fi, ei;

  fprintf( fout, "%d %d\n", n_vertex, n_face);
  for (i=0; i<n_vertex; i++)
    {
    fprintf( fout, "%20.10lf %20.10lf %20.10lf\n", list_v[i].V[0], list_v[i].V[1], list_v[i].V[2]);
    }
  for (fi=0; fi<n_face; fi++)
    {
    fprintf( fout, "%d ", list_f[fi].nVert);
    for (ei=0; ei<list_f[fi].nVert; ei++)
      fprintf( fout, "%d ", list_f[fi].V[ei]);
    for (ei=0; ei<list_f[fi].nVert; ei++)
      fprintf( fout, "%d ", list_f[fi].F[ei]);
    fprintf( fout, "\n");
    }
  }

void InsertEdge( Edge **list_e, int e1, int e2, int fi)
  {
  int t, Side;
  Edge *e;

  Side = LEFT;

  if (e1 > e2)
    {
    t = e1;
    e1 = e2;
    e2 = t;
    Side = RIGHT;
    }

  e = list_e[e1];

  while (e!= NULL)
    {
    if (e->V[1] == e2)
      {
      if (e->F[Side] != -1)
        {
        fprintf( stderr, "More than 2 faces meeting along one edge!!!\n");
        fprintf( stderr, "Edge %d -> %d: Face %d, %d and %d!\n", e->V[0], e->V[1], e->F[0], e->F[1], fi);
        exit(-1);
        }

      fprintf( stderr, "Adding Face %03d to Edge %03d-%03d\n", fi, e1, e2);

      e->F[Side] = fi;
      return;
      }
    e = e->Next;
    }

  if (e == NULL)
    {
    e = (Edge*) malloc(sizeof(Edge));
    e->Next = list_e[e1];
    list_e[e1] = e;
    e->V[0] = e1;
    e->V[1] = e2;
    e->F[Side] = fi;
    e->F[1-Side] = -1;

    fprintf( stderr, "Create Face %03d at Edge %03d-%03d\n", fi, e1, e2);

    }
  }

void Off2Edge( int n_vertex, int n_face, Vertex *list_v, Face *list_f, Edge **list_e)
  {
  int fi, ei;

  for ( fi=0; fi<n_face; fi++)
    for ( ei=0; ei<list_f[fi].nVert; ei++)
            {
            fprintf( stderr, "Inserting in face %03d edge %03d-%03d\n", fi, list_f[fi].V[ei], list_f[fi].V[(ei+1)%list_f[fi].nVert]);
	    InsertEdge( list_e, list_f[fi].V[ei], list_f[fi].V[(ei+1)%list_f[fi].nVert], fi);
            }
  }

/*---------------------------------------------*/
/* RETURN THE INDEX OF VERTEX v IN FACE list_f */
/*---------------------------------------------*/

int IndexPointOnFace( Face *list_f, int v)
  {
  int ei;

  for (ei=0; ei<list_f->nVert; ei++)
    if (list_f->V[ei] == v) return ei;

  fprintf( stderr, "Erreur topologique\n");
  exit(-1);
  return -1;
  }

void ConnectFaces( int n_vertex, Face *list_f, Edge **list_e)
  {
  int v;
  Edge *e;
  int e1;

  for (v=0; v<n_vertex; v++)
    {
    e = list_e[v];

    while (e != NULL)
      {
      if (e->F[LEFT] != -1)
        {
        e1 = IndexPointOnFace( &list_f[e->F[LEFT]], e->V[0]);
        if (list_f[e->F[LEFT]].V[(e1+1) % list_f[e->F[LEFT]].nVert] != e->V[1])
          {
          fprintf( stderr, "Edge %d %d should have Face %d on the Left\n", e->V[0], e->V[1], e->F[LEFT]);
          exit(-1);
          }
        else
          {
          list_f[e->F[LEFT]].F[e1] = e->F[RIGHT];
          fprintf( stderr, "Connecting faces %03d and %03d\n", e->F[LEFT], e->F[RIGHT]);
          }
        }
      if (e->F[RIGHT] != -1)
        {
        e1 = IndexPointOnFace( &list_f[e->F[RIGHT]], e->V[1]);
        if (list_f[e->F[RIGHT]].V[(e1+1) % list_f[e->F[RIGHT]].nVert] != e->V[0])
          {
          fprintf( stderr, "Edge %d %d should have Face %d on the Right\n", e->V[0], e->V[1], e->F[RIGHT]);
          exit(-1);
          }
        else
          {
          list_f[e->F[RIGHT]].F[e1] = e->F[LEFT];
          fprintf( stderr, "Connecting faces %03d and %03d\n", e->F[RIGHT], e->F[LEFT]);
          }
        }
      e = e->Next;
      }
    }
  }

int main( int argc, char *argv[])
  {
  int v, n_vertex, n_face, Len;
  Vertex *list_v;
  FILE *fin, *fout;
  Face *list_f;
  Edge **list_e;

  if (argc != 2) {
	  fprintf( stderr, "Usage: %s <OFF filename>\n", argv[0]);
	  exit(-1);
  }

  if ((fin = fopen( argv[1], "r")) == 0) {
	  fprintf( stderr, "Unable to read file %s\n", argv[1]);
  }

  ReadOff( fin, &n_vertex, &n_face, &list_v, &list_f);

  fclose(fin);

  Len = strlen( argv[1]);

  argv[1][Len-3] = 'g';
  argv[1][Len-2] = argv[1][Len];

  fout = fopen( argv[1], "w");

  list_e = (Edge**) malloc(sizeof(Edge*) * n_vertex);

  if (list_e == NULL) {PrintError( MEMORY, "edges");}

  for (v=0; v<n_vertex; v++) list_e[v] = NULL;

  Off2Edge( n_vertex, n_face, list_v, list_f, list_e);

  ConnectFaces( n_vertex, list_f, list_e);

  WriteGFile( fout, n_vertex, n_face, list_v, list_f);

  fclose(fout);
  }
