#include <stdio.h>
#include <stdlib.h>

typedef double Point3D[3];

typedef struct s_Face{
	int nArete;         /* nombre d'artes de la face */
	int *S;             /* tableau des indices des sommets de la face */
	int *FV;            /* tableau des indices des faces voisines */
	int *IndAr;         /* tableau temporaire des indices absolu des artes de la face */
	int IndFa;          /* premier indice des nouvelles faces aprs l'oprateur MD */
} Face;

typedef struct s_Grille{
	int nSommet, nFace; /* nombre de sommets et de face dans la grille */
	Point3D *S;         /* tableau des coordonnes des sommets */
	Face *F;            /* tableau des faces */
} Grille;

/* Calcul le barycentre B de la face indice fi dans la grille G */
void Barycentre( Grille *G, int fi, Point3D B)
{
	int i,IndS;
	double Coeff;

	B[0] = 0.;
	B[1] = 0.;
	B[2] = 0.;

	for (i=0; i<G->F[fi].nArete; i++) {
		IndS = G->F[fi].S[i];

		B[0] += G->S[IndS][0];
		B[1] += G->S[IndS][1];
		B[2] += G->S[IndS][2];
	}

	Coeff = 1. / G->F[fi].nArete;
	B[0] *= Coeff;
	B[1] *= Coeff ;
	B[2] *= Coeff;
}

void LiberationGrille( Grille *Sortie)
{
	int fi;

	free(Sortie->S);

	for (fi=0; fi<Sortie->nFace; fi++) {
		if (Sortie->F[fi].nArete != -1) {
			free(Sortie->F[fi].S);
			free(Sortie->F[fi].FV);
		}
	}

	if (Sortie->nFace > 0) {
		free(Sortie->F);
	}

	free(Sortie);
}

void LectureGrille( char *NomDeFichier, Grille **pSortie)
{
	FILE *f;
	Grille *Sortie;
	int si, fi;

	Sortie = (Grille*) malloc(sizeof(Grille));
	*pSortie = Sortie;

	f = fopen ( NomDeFichier, "r");

	fscanf( f, "%d", &Sortie->nSommet);
	fscanf( f, "%d", &Sortie->nFace);

	fprintf( stderr, "Nbre de sommets en entre = %d\n", Sortie->nSommet);
	fprintf( stderr, "Nbre de faces en entre = %d\n", Sortie->nFace);

	Sortie->S = (Point3D*) malloc(sizeof(Point3D)*Sortie->nSommet);
	Sortie->F = (Face*) malloc(sizeof(Face)*Sortie->nFace);

	for (si=0; si<Sortie->nSommet; si++) {
		fscanf( f, "%lf", &Sortie->S[si][0]);
		fscanf( f, "%lf", &Sortie->S[si][1]);
		fscanf( f, "%lf", &Sortie->S[si][2]);
	}

	for (fi=0; fi<Sortie->nFace; fi++) {
		fscanf( f, "%d", &Sortie->F[fi].nArete);

		Sortie->F[fi].S = (int*) malloc(sizeof(int) * Sortie->F[fi].nArete);
		Sortie->F[fi].FV = (int*) malloc(sizeof(int) * Sortie->F[fi].nArete);

		for (si=0; si<Sortie->F[fi].nArete; si++) {
			fscanf( f, "%d", &Sortie->F[fi].S[si]);
		}

		for (si=0; si<Sortie->F[fi].nArete; si++) {
			fscanf( f, "%d", &Sortie->F[fi].FV[si]);
		}
	}
}

void EcritureGrille( char *NomDeFichier, Grille *Entree)
{
	FILE *f;
	int si, fi;

	f = fopen( NomDeFichier, "w");

	fprintf( f, "%d %d\n", Entree->nSommet, Entree->nFace);

	for (si=0; si<Entree->nSommet; si++)
		fprintf( f, "%10.5lf %10.5lf %10.5lf\n", Entree->S[si][0], Entree->S[si][1], Entree->S[si][2]);

	for (fi=0; fi<Entree->nFace; fi++) {
		fprintf( f, "%d ", Entree->F[fi].nArete);

		for (si=0; si<Entree->F[fi].nArete; si++)
			fprintf( f, "%d ", Entree->F[fi].S[si]);

		for (si=0; si<Entree->F[fi].nArete; si++)
			fprintf( f, "%d ", Entree->F[fi].FV[si]);

		fprintf( f, "\n");
	}

	fclose(f);
}

void EcritureGrilleVRML( char *NomDeFichier, Grille *Entree)
{
	FILE *f;
	int si, fi;

	f = fopen( NomDeFichier, "w");

	fprintf( f, "#VRML V1.0 ascii\n");

	fprintf( f, "Separator {\n");
	fprintf( f, "   Separator {\n");
	fprintf( f, "        Coordinate3 {\n");
	fprintf( f, "            point [\n");

	for (si=0; si<Entree->nSommet-1; si++) {
		fprintf( f, "%10.5lf %10.5lf %10.5lf ,\n",
				Entree->S[si][0],
				Entree->S[si][1],
				Entree->S[si][2]);
	}
	fprintf( f, "%10.5lf %10.5lf %10.5lf]}\n",
			Entree->S[si][0],
			Entree->S[si][1],
			Entree->S[si][2]);
	fprintf( f, "Separator {\n");
	fprintf( f, "	    IndexedFaceSet {\n");
	fprintf( f, "			coordIndex	[ \n");

	for (fi=0; fi<Entree->nFace-1; fi++) {
		for (si=0; si<Entree->F[fi].nArete; si++)
			fprintf( f, "%d, ", Entree->F[fi].S[si]);
		fprintf( f, "-1,\n");
		}

	if (Entree->nFace > 0) {
		for (si=0; si<Entree->F[fi].nArete; si++)
			fprintf( f, "%d, ", Entree->F[fi].S[si]);
	}
	fprintf( f, "-1]\n");
	fprintf( f, "}}}}\n");

	fclose(f);
}

int RechercheSommetAbsolu( Grille *G, int fi, int SommetAbsolu)
{
	int si;

	si = 0;

	while (si < G->F[fi].nArete) {
		if (G->F[fi].S[si] == SommetAbsolu) return si;
		si = si+1;
	}

	fprintf( stderr, "Erreur de topologie dans RechercheSommetAbsolu!\n");
	return -1;
}

void OperateurMD( Grille *Entree, Grille **pSortie)
{
	Grille *Sortie;
	int IndRel, nArete, nFace, fi, si, fj, Ind1, Ind2, IndSor, IndFa;

	Sortie = (Grille*) malloc(sizeof(Grille));
	*pSortie = Sortie;

	/* On parcours les artes et on calcul les milieux des artes */

	nArete = 0;
	nFace = 0;
	for (fi=0; fi<Entree->nFace; fi++) {

		Entree->F[fi].IndAr = (int*) malloc(sizeof(int) * Entree->F[fi].nArete);
		Entree->F[fi].IndFa = nFace;
		nFace += Entree->F[fi].nArete;

		for (si=0; si<Entree->F[fi].nArete; si++) {

			fj = Entree->F[fi].FV[si];

	/* On ne traite l'arte entre fi et fj que si fi > fj */

			if (fi>fj) {
				Entree->F[fi].IndAr[si] = nArete;
				if (fj != -1 ) {
					IndRel = RechercheSommetAbsolu( Entree, fj, Entree->F[fi].S[(si+1) % Entree->F[fi].nArete]);
					Entree->F[fj].IndAr[IndRel] = nArete;
				}

				nArete++;	
			}
		}

	}

//	fprintf( stderr, "Oprateur MD: Nbre d'artes = %d\n", nArete);
//	fprintf( stderr, "Oprateur MD: Nbre de nouvelles faces = %d\n", nFace);

	Sortie->nSommet = Entree->nSommet + Entree->nFace + nArete;
	Sortie->S = (Point3D*) malloc(sizeof(Point3D) * Sortie->nSommet);

	Sortie->nFace = nFace;
	if (nFace != 0) { Sortie->F = (Face*) malloc(sizeof(Face) * nFace); }

	/* On recopie les anciens sommets */
	for (si=0; si<Entree->nSommet; si++) {
		Sortie->S[si][0] = Entree->S[si][0];
		Sortie->S[si][1] = Entree->S[si][1];
		Sortie->S[si][2] = Entree->S[si][2];
	}

	/* On calcule les barycentres des anciennes faces */
	for (fi=0; fi<Entree->nFace; fi++) {
		Barycentre( Entree, fi, Sortie->S[Entree->nSommet + fi]);
	}

	/* On calcule les milieux des anciennes artes */
	for (fi=0; fi<Entree->nFace; fi++) {
		for (si=0; si<Entree->F[fi].nArete; si++) {

			fj = Entree->F[fi].FV[si];

			if (fi>fj) {
				Ind1 = Entree->F[fi].S[si];
				Ind2 = Entree->F[fi].S[(si+1) % Entree->F[fi].nArete];

				IndSor = Entree->F[fi].IndAr[si] + Entree->nSommet + Entree->nFace;

				Sortie->S[IndSor][0] = 0.5 * (Entree->S[Ind1][0] + Entree->S[Ind2][0]);
				Sortie->S[IndSor][1] = 0.5 * (Entree->S[Ind1][1] + Entree->S[Ind2][1]);
				Sortie->S[IndSor][2] = 0.5 * (Entree->S[Ind1][2] + Entree->S[Ind2][2]);
			}
		}
	}

	/* On initialise les nouvelles faces */
	for (fi=0; fi<Entree->nFace; fi++) {

		nArete = Entree->F[fi].nArete;
		for (si=0; si<nArete; si++) {

			IndFa = Entree->F[fi].IndFa + si;
			Sortie->F[IndFa].nArete = 4;
			Sortie->F[IndFa].S = (int*) malloc(sizeof(int) * 4);
			Sortie->F[IndFa].FV = (int*) malloc(sizeof(int) * 4);

	/* Indice du barycentre de la face */
			Sortie->F[IndFa].S[0] = Entree->nSommet + fi;
	/* Indice du milieu de la premire ancienne arte */
			Sortie->F[IndFa].S[1] = Entree->nSommet + Entree->nFace + Entree->F[fi].IndAr[(si-1+nArete) % nArete];
	/* Indice de l'ancien sommet */
			Sortie->F[IndFa].S[2] = Entree->F[fi].S[si];
	/* Indice du milieu de la seconde ancienne arte */
			Sortie->F[IndFa].S[3] = Entree->nSommet + Entree->nFace + Entree->F[fi].IndAr[si];
		}
	}

	/* On connecte les nouvelles faces */
	for (fi=0; fi<Entree->nFace; fi++) {

		nArete = Entree->F[fi].nArete;
		for (si=0; si<nArete; si++) {

			IndFa = Entree->F[fi].IndFa;

			Sortie->F[IndFa+si].FV[0] = IndFa + ((si + nArete - 1) % nArete);

			fj = Entree->F[fi].FV[(si + nArete - 1) % nArete];
			if (fj != -1) {
				IndRel = RechercheSommetAbsolu( Entree, fj, Entree->F[fi].S[si]);
				Sortie->F[IndFa+si].FV[1] = Entree->F[fj].IndFa + IndRel;
			} else {
				Sortie->F[IndFa+si].FV[1] = -1;
			}

			fj = Entree->F[fi].FV[si];
			if (fj != -1) {
				IndRel = RechercheSommetAbsolu( Entree, fj, Entree->F[fi].S[si]);
				Sortie->F[IndFa+si].FV[2] = Entree->F[fj].IndFa + IndRel;
			} else {
				Sortie->F[IndFa+si].FV[2] = -1;
			}

			Sortie->F[IndFa+si].FV[3] = IndFa + ((si + 1) % nArete);
		}
	}

	/* On libre la place du tableau temporaire contenant les indices d'artes */
	for (fi=0; fi<Entree->nFace; fi++)
		free( Entree->F[fi].IndAr);
}

void OperateurM( Grille *Entree, Grille **pSortie)
{
	Grille *Sortie, *VraiSortie;
	int fi, si, sj, nA, NouvelleFace, SommetAbsolu, nANouvelleFace;
	int I1, I2, IndRel1, IndRel2, FaceARechercher;
	int IndCourant, *IndFace;

	#define MAX 100
	int SNV[MAX];

	Sortie = (Grille*) malloc(sizeof(Grille));

	/* Le nombre de nouvelles faces est gal au nombre d'anciens sommets internes */
	/* donc le nombre de nouvelles faces est plus petit que le nombre d'anciens sommets */
	
	Sortie->nFace = 0; /* on initialise  zro, avant de connaitre le nbre de nouvelles faces */
	if (Entree->nSommet > 0) {
		Sortie->F = (Face*) malloc(sizeof(Face) * Entree->nSommet);
	}
	for (si=0; si<Entree->nSommet; si++)
		Sortie->F[si].nArete = -2; /* Nouvelle face pas encore visite */

	/* Les nouveaux sommets sont les barycentres des anciennes faces */

	Sortie->nSommet = Entree->nFace;

	Sortie->S = (Point3D*) malloc(sizeof(Point3D) * Sortie->nSommet);

	for (fi=0; fi<Sortie->nSommet; fi++)
		Barycentre( Entree, fi, Sortie->S[fi]);

	/* On tourne autour des anciens sommets pour calculer les nouvelles faces */

	for (fi=0; fi<Entree->nFace; fi++) {

		for (si=0; si<Entree->F[fi].nArete; si++) {

			SommetAbsolu = Entree->F[fi].S[si];

			if (Sortie->F[SommetAbsolu].nArete != -2) {
//				fprintf( stderr, "Sommet dj visit!\n");
			} else {

//				fprintf( stderr, "Je tourne autour du sommet %d\n", SommetAbsolu);

//				fprintf( stderr, "Je rencontre la face %d\n", fi);
				SNV[0] = fi;

				nANouvelleFace = 1;

				nA = Entree->F[fi].nArete;
				NouvelleFace = Entree->F[fi].FV[(si+nA-1) % nA];

				while ((NouvelleFace != fi) && (NouvelleFace != -1)) {

//					fprintf( stderr, "Je rencontre la face %d\n", NouvelleFace);
					SNV[nANouvelleFace] = NouvelleFace;

					sj = RechercheSommetAbsolu( Entree, NouvelleFace, SommetAbsolu);
					nA = Entree->F[NouvelleFace].nArete;
					NouvelleFace = Entree->F[NouvelleFace].FV[(sj+nA-1) % nA];


					nANouvelleFace++;
				}
				if (NouvelleFace == -1) {
	/* on est tomb sur un sommet de bord => on ne cre pas de nouvelle face */
					Sortie->F[SommetAbsolu].nArete = -1;
				} else {
	/* cration de la nouvelle face */
					Sortie->nFace++;

					Sortie->F[SommetAbsolu].nArete = nANouvelleFace;
					Sortie->F[SommetAbsolu].S = (int*) malloc(sizeof(int) * nANouvelleFace);
					Sortie->F[SommetAbsolu].FV = (int*) malloc(sizeof(int) * nANouvelleFace);

					for (sj=0; sj<nANouvelleFace; sj++)
						Sortie->F[SommetAbsolu].S[sj]=SNV[sj];

//					fprintf( stderr,"le sommet %d de la face %d est interne, et d'ordre %d\n", si, fi, nANouvelleFace);
				}
			}
		}
	}

//	fprintf( stderr, "Nbre de nouvelles face aprs oprateur M = %d\n", Sortie->nFace);
	if (Sortie->nFace == 0) {
		fprintf( stderr, "Attention: il n'y a plus de face dans la grille!\n");
	}

	/* On tourne autour des anciennes faces pour connecter les nouvelles faces */

	for (fi=0; fi<Entree->nFace; fi++) {

		for (si=0; si<Entree->F[fi].nArete; si++) {

	/* I1 et I2 sont les indices des nouvelles faces  connecter */

			I1 = Entree->F[fi].S[si];

			if (Sortie->F[I1].nArete == -1) {I1 = -1;}

			I2 = Entree->F[fi].S[(si+1) % Entree->F[fi].nArete];

			if (Sortie->F[I2].nArete == -1) {I2 = -1;}

	/* On recherche dans la nouvelle face d'indice I1 o se trouve l'ancienne face en face de fi */

			FaceARechercher=Entree->F[fi].FV[si];

			if (FaceARechercher != -1) {
				if (I1 != -1) {
					IndRel1 = RechercheSommetAbsolu( Sortie, I1, FaceARechercher);
				}
				if (I2 != -1) {
					IndRel2 = RechercheSommetAbsolu( Sortie, I2, FaceARechercher);
					IndRel2 = (IndRel2 - 1 + Sortie->F[I2].nArete) % Sortie->F[I2].nArete;
				}

				if (I1 != -1) { Sortie->F[I1].FV[IndRel1] = I2; }
				if (I2 != -1) { Sortie->F[I2].FV[IndRel2] = I1; }
			}
		}
	}

//	fprintf( stderr, "Nbre de faces aprs oprateur M = %d\n", Sortie->nFace);

	/* On supprime les faces dont le nbre d'artes est  -1 */

	IndFace = (int*) malloc(sizeof(int) * Entree->nSommet);

	IndCourant = 0;
	for (fi=0; fi<Entree->nSommet; fi++) {

		if (Sortie->F[fi].nArete != -1) {
			IndFace[fi] = IndCourant;

//			fprintf( stderr, "IndFace[%d] = %d\n", fi, IndCourant);
			IndCourant++;
		}
	}

	VraiSortie = (Grille*) malloc(sizeof(Grille));
	*pSortie = VraiSortie;

	VraiSortie->nFace = Sortie->nFace;
	VraiSortie->nSommet = Sortie->nSommet;

	VraiSortie->S = (Point3D*) malloc(sizeof(Point3D) * VraiSortie->nSommet);
	if (Sortie->nFace > 0) {
		VraiSortie->F = (Face*) malloc(sizeof(Face) * VraiSortie->nFace);
	}

	for (si=0; si<VraiSortie->nSommet; si++) {
		VraiSortie->S[si][0] = Sortie->S[si][0];
		VraiSortie->S[si][1] = Sortie->S[si][1];
		VraiSortie->S[si][2] = Sortie->S[si][2];
	}

	IndCourant = 0;
	for (fi=0; fi<Entree->nSommet; fi++) {

		if ( (Sortie->F[fi].nArete != -1) && (Sortie->F[fi].nArete != -2) ) {

			VraiSortie->F[IndCourant].nArete = Sortie->F[fi].nArete;
			VraiSortie->F[IndCourant].S = (int*) malloc(sizeof(int) * Sortie->F[fi].nArete);
			VraiSortie->F[IndCourant].FV = (int*) malloc(sizeof(int) * Sortie->F[fi].nArete);

			for (si=0; si<Sortie->F[fi].nArete; si++) {

				VraiSortie->F[IndCourant].S[si] = Sortie->F[fi].S[si];
				if (Sortie->F[fi].FV[si] == -1) {
					VraiSortie->F[IndCourant].FV[si] = -1;
				} else {
					VraiSortie->F[IndCourant].FV[si] = IndFace[ Sortie->F[fi].FV[si] ];
				}
			}
			IndCourant++;
		}
	}

	free(IndFace);
	LiberationGrille(Sortie);
}

int main(int argc, char *argv[])
{
	int n, d;
	char NomFichier[100];
	Grille *Grille1, *Grille2;

	if (argc != 4) {
		fprintf( stderr, "usage: subdivision <nom de fichier> <degre> <nbre d'etapes de subdivision>\n");
		fprintf( stderr, "subdivision cube.g 3 5 => subdivision du cube, pour le degre 3, avec 5 etapes de subdivision\n");
		exit(-1);
	}

	LectureGrille( argv[1], &Grille1);

	sprintf( NomFichier, "sortie00.wrl");
	EcritureGrilleVRML( NomFichier, Grille1);

	for (n = 0; n<atoi(argv[3]); n++) {
		OperateurMD( Grille1, &Grille2);
		LiberationGrille( Grille1);

		for (d=0; d<atoi(argv[2])-1; d++) {
			OperateurM( Grille2, &Grille1);
			LiberationGrille( Grille2);
			Grille2 = Grille1;
		}

		fprintf( stderr, "Nbre de faces aprs %d subdivisions: %d\n", n+1, Grille2->nFace);
		if (Grille2->nFace > 0) {
			sprintf( NomFichier, "sortie%02d.wrl", n+1);
			EcritureGrilleVRML( NomFichier, Grille2);
		}
		fprintf( stderr, "Fichier crit: %s\n", NomFichier);
		Grille1 = Grille2;
	}
}
